"""
    To install MobsPy just use pip:
    pip install mobspy

    Another option is to download the github and use
    pip install -e 'downloaded file location'
"""

# Import MobsPy
from mobspy import *
"""
    To start off let us simulate a simple A + B >> 2*C + D reaction
"""

# First, we define the meta-species that will be used in the reaction.
# We use the BaseSpecies constructor with the number of desired meta-species.
A, B, C, D = BaseSpecies(4)

# Now we write the reaction. It follows the CRN syntax
# >> operator:
#   With the >> operator separating reactants on the left and products
#   on the right.
# Initial counts:
#   Initial counts are assigned using the call operator.
#   For example A(200) initializes species A with count 200.
# Reaction rate:
#   Each reaction has a rate that is assigned using the [] operator.
#   If a float is given as a rate, we assume mass action kinetics with the
#   float being the reaction rate constant.
A(200) + B(100) >> 2*C + D [0.1]

# This will create a reaction A + B -> 2*C + D
# with a rate equal to [A]*[B]*0.1.

# We next construct the Simulation object by giving it the meta-species we want to simulate.
# Species are named according to their variables.
My_Sim = Simulation(A | B | C | D)

# To configure the simulation just use the dot operator with the
# respective attribute.
# The available attributes are documented in a read_me in the parameter folder.
My_Sim.save_data = False
My_Sim.duration = 5
My_Sim.volume = 10
My_Sim.simulation_method = 'stochastic'

# Finally the simulation is run and results
# printed via
My_Sim.run()



import sys, os
import matplotlib.pyplot as plt

from mobspy import *

"""
    Here we have the implementation of a toggle switch
    Since MobsPy does not yet have an event implementation, we call the model again with changes made to it
    In the future we will implement a way to handle events, but for now events can be handled in a similar fashion
"""

"""
    We currently access a private variable to implement event
    Since we are planning on implementing a better syntax for event-handling
    In future updates expect a better way to handle this
    But for now this method is an alternative
"""

p1 = []
p2 = []
T = []


# Toggle Switch
def set_model_initial_conditions(simulation, state_dict):
    simulation.compile(verbose=False)
    for key in state_dict:
        for spe in simulation._species_for_sbml:
            value = state_dict[key]
            species_state = key.replace('.', '_dot_')
            if species_state == spe:
                simulation._species_for_sbml[species_state] = value


def event_conditions(simulation, event_dict):
    for key in event_dict:
        for spe in simulation._species_for_sbml:
            value = event_dict[key]
            species_state = key.replace('.', '_dot_')
            if species_state == spe:
                simulation._species_for_sbml[species_state] = value


def toggle_switch(g1_count=0, g2_count=0, state_dict={}, duration=500000):
    Promoter, Cas_binding, dCas9 = BaseSpecies(3)
    Cas_binding.cas, Cas_binding.no_cas
    Promoter.inactive, Promoter.active
    gRNA = New(Cas_binding, 1)
    DNAPromoter = New(Promoter, 1)
    Promoter_list = ['P1', 'P2']
    gRNA_list = ['g1', 'g2']

    # g_RNA pdn
    for grna, prom in zip(gRNA_list, Promoter_list):
        DNAPromoter.active.c(prom) >> gRNA.no_cas.c(grna) + DNAPromoter.active.c(prom)[0.0833]

    # Degradation rxns
    gRNA.no_cas >> Zero[0.0069]
    Rev[Zero >> dCas9][1 / 60, 0.000383]

    # Promoter Pdn and Degradation
    Promoter.active >> 2 * Promoter.active[0.000383]
    Promoter >> Zero[0.000383]

    # gRNA:dCas9 Complex Formation
    Rev[gRNA.no_cas + dCas9 >> gRNA.cas][0.0018 * 6.022e-1, 0.000383]

    # Complex formation at the promoter and Complex falloff during Repn
    gRNA_rep_list = ['g2', 'g1']
    for grna, prom in zip(gRNA_rep_list, Promoter_list):
        DNAPromoter.active.c(prom) + gRNA.cas.c(grna) >> DNAPromoter.inactive.c(prom)[0.012 * 6.022e-1]
        DNAPromoter.inactive.c(prom) >> 2 * DNAPromoter.active.c(prom) + gRNA.cas.c(grna)[0.000383]

    MySim = Simulation(gRNA | DNAPromoter | dCas9)
    MySim.save_data = False
    MySim.plot_data = False
    MySim.duration = duration
    if not state_dict:
        DNAPromoter.P1.active(1), DNAPromoter.P2.active(1)
        gRNA.no_cas.g1(g1_count), gRNA.no_cas.g2(g2_count)
        dCas9(2)
        MySim.run()
    else:
        if g1_count != 0:
            event_dict = {'gRNA.no_cas.g1': g1_count}
        if g2_count != 0:
            event_dict = {'gRNA.no_cas.g2': g2_count}
        set_model_initial_conditions(MySim, state_dict)
        event_conditions(MySim, event_dict)
        MySim.run(compile=False)

    state_dict = {}
    for key in MySim.results['data']:
        if key == 'Time':
            pass
        else:
            state_dict[key] = MySim.results['data'][key]['runs'][0][-1]

    global p1
    global p2
    p1 += MySim.results['data']['DNAPromoter.P1.active']['runs'][0]
    p2 += MySim.results['data']['DNAPromoter.P2.active']['runs'][0]
    return state_dict


n = 1e10
state = toggle_switch(g1_count=10 / 6.022e-1, g2_count=0)
state = toggle_switch(g1_count=0, g2_count=n*10 / 6.022e-1, state_dict=state)
state = toggle_switch(g1_count=n*10/6.022e-1, g2_count=0, state_dict=state)
state = toggle_switch(g1_count=0, g2_count=n*10/6.022e-1, state_dict=state)
plt.plot(p1)
plt.plot(p2)
plt.show()




"""
    This module is responsible for the construction of all individual reactions from a meta-reaction
"""
from copy import deepcopy
import mobspy.modules.function_rate_code as fr
import itertools
import mobspy.modules.meta_class as mc
import mobspy.simulation_logging.log_scripts as simlog
import mobspy.modules.species_string_generator as ssg
from inspect import signature
from mobspy.modules.order_operators import Default


def iterator_for_combinations(list_of_lists):
    """
        Iterate through all the combinations of a list of lists
        [[A,B,C], [D, F, E], [G]] means:
        ADG, AFG, AEG, BDG, BFG, BEG, CDG, CFG, CEG .....

        Parameter:
            list_of_lists (list of lists) = list of all lists to iterate through all combinations
    """
    for i in itertools.product(*list_of_lists):
        yield i


def copy_reaction(reaction):
    """
        Just copies a meta-reaction
        Deepcopy is not working because it calls __getattr__ on the species with a private method

        Parameters:
            reaction (meta-reaction) = meta-reaction to be copied
    """
    reactants = []
    for reactant in reaction.reactants:
        characteristics = deepcopy(reactant['characteristics'])
        species = reactant['object']
        stoichiometry = reactant['stoichiometry']
        reactants.append({'object': species, 'characteristics': characteristics,
                          'stoichiometry': stoichiometry})

    products = []
    for product in reaction.products:
        characteristics = deepcopy(product['characteristics'])
        species = product['object']
        stoichiometry = product['stoichiometry']
        products.append({'object': species, 'characteristics': characteristics,
                         'stoichiometry': stoichiometry})

    reaction_copy = mc.Reactions(reactants, products)
    reaction_copy.set_rate(reaction.rate)
    return reaction_copy


def check_for_invalid_reactions(reactions, ref_characteristics_to_object):
    """
        If query references two independent characteristics from the in meta-species it would result in an empty set
        Like Ecoli.live.dead - (assuming live and dead belong to the same meta-species characteristics set)
        If that ever happens inside a meta-reaction we just pop an error

        :param reactions: (set of meta-reactions) set of meta-reactions inside the model
        :param ref_characteristics_to_object: (dict) dictionary with the characteristics as keys and their respective
        object as values
    """
    for reaction in reactions:
        for reactant in reaction.reactants:

            check_for_duplicates = {}
            for cha in reactant['characteristics']:

                # Ok I love try catches for checking if something is in a dict. Don't judge me
                try:
                    check_for_duplicates[ref_characteristics_to_object[cha]]
                    simlog.error(f'Illegal reaction: {reaction}. \n'
                                 'There is a query with two characteristics '
                                 f'{ref_characteristics_to_object[cha].get_characteristics()} from the same'
                                 f' vector axis resulting in an impossible query. \n'
                                 f'As all those characteristics have been directly added to'
                                 f'{ref_characteristics_to_object[cha]}, they are located in the same vector axis. \n'
                                 f'To solve this either assign the characteristics to new meta-species and use '
                                 f'inheritance or create a new reaction for each desired query.')
                except KeyError:
                    try:
                        check_for_duplicates[ref_characteristics_to_object[cha]] = cha
                    except KeyError:
                        simlog.error(
                            'The characteristic\'s object was not found in the species supplied to the simulator \n'
                            'Perhaps a species is missing ?')

        for product in reaction.products:

            check_for_duplicates = {}
            for cha in product['characteristics']:

                # Ok I love try catches for checking if something is in a dict. Don't judge me
                try:
                    check_for_duplicates[ref_characteristics_to_object[cha]]
                    simlog.error(f'Illegal reaction: {reaction}. \n'
                                 'There is a transformation with two characteristics '
                                 f'{ref_characteristics_to_object[cha].get_characteristics()} from the same'
                                 f' vector axis resulting in an undefinable transformation. \n'
                                 f'As all those characteristics have been directly added to '
                                 f'{ref_characteristics_to_object[cha]}, they are located in the same vector axis. \n'
                                 f'To solve this either assign the characteristics to new meta-species and use '
                                 f'inheritance or create a new reaction for each desired query.')
                except KeyError:
                    if '$' not in cha:
                        check_for_duplicates[ref_characteristics_to_object[cha]] = cha


def construct_reactant_structures(reactant_species, ref_characteristics_to_object):
    """
        This function finds the corresponding strings according to the reactant meta-species with or without a query
        pack then in a list and return

        :param reactant_species: (meta-species object) species objects of the involved species
        :param ref_characteristics_to_object: (dict) dictionary with characteristics as keys and species objects as
        values
    """
    species_string_combinations = []

    for reactant in reactant_species:
        species_string_combinations.append(ssg.construct_all_combinations(reactant['object'],
                                                                          reactant['characteristics'],
                                                                          ref_characteristics_to_object))

    return species_string_combinations


def construct_order_structure(species_order_list, current_species_string_list):
    """
        Order structure for reaction order operations. Returns the cyclic_dictionary to be used by the order operator.
        The meta-species objects are the keys of this dictionary and a lists of species strings currently being used
        in the reaction are the values - Allowing the product to find it's corresponding species-string in a future
        step

        :param species_order_list: (list of meta-species objects) list of meta-species objects as they appear in the
        meta-reaction
        :param current_species_string_list: (list of strings) list of strings in MobsPy format of the species
        currently in this specific reaction

        :return: cyclic_dict (dict) Dictionary where the keys are meta-species objects and the values are
        lists of species
    """
    cyclic_dict = {}
    for species_object, species_string in zip(species_order_list, current_species_string_list):

        try:
            cyclic_dict[species_object].append(species_string)
        except KeyError:
            cyclic_dict[species_object] = [species_string]

    return cyclic_dict


def construct_product_structure(reaction):
    """
        This function unpacks the products in a meta-reaction

        :param: reaction meta-reaction currently being analysed

        :return: product_list = A list of dictionaries for each product with the meta-species object, the label and the
        characteristics
    """
    product_list = []
    for product in reaction.products:
        for _ in range(product['stoichiometry']):
            product_list.append({'species': product['object'], 'label': product['label'],
                                 'characteristics': product['characteristics']})

    return product_list


def construct_single_reaction_for_sbml(reactant_species_string_list, product_species_string_list, reaction_rate):
    """
        This function constructs the reactions for SBML for the conversion by the model builder script
        It follows the following structure 're':[('stoichmetry', reactantant_string) ....
        The reaction rate must be a string containing the reaction kinetics
        This returns a single reaction to be appended by the reactions_for_sbml dictionary

        :param reactant_species_string_list: (list of strings) list of reactants in MobsPy format
        :param product_species_string_list: (list of strings) list of products in MobsPy format
        :param reaction_rate: (str) reaction rate expression as a string

        :return: to_return (dict) = dictionary that packs the reactants products and rate
    """
    to_return = {'re': [], 'pr': [], 'kin': reaction_rate}

    reactant_count_dict = count_string_dictionary(reactant_species_string_list)
    product_count_dict = count_string_dictionary(product_species_string_list)

    for key in reactant_count_dict:
        to_return['re'].append((reactant_count_dict[key], key))

    for key in product_count_dict:
        to_return['pr'].append((product_count_dict[key], key))

    return to_return


def count_string_dictionary(list_of_strings):
    """
        Count the number of instances in a list and return them in a dictionary where the keys are the strings and
        the value the number of times it appeared in the list
    """
    to_return = {}

    for e in list_of_strings:
        try:
            to_return[e] += 1
        except KeyError:
            to_return[e] = 1

    return to_return


def get_involved_species(reaction, meta_species_in_model):
    """
        This extracts all the involved meta-species inside a reaction
        This function is responsible for implementing the inheritance mechanism, by finding within each meta-species
        references set, if they reference the meta-species in the reaction

        :param reaction: (meta-reaction object)
        :param meta_species_in_model: (list) list of meta-species used in the model

        :return: base_species_order (list of meta-species objects) = order that the meta-species appear in the
        meta-reaction, reactant_species_combination_list (list of lists of meta-species) = list of lists of all
        meta-species that have inherited from the meta-species in the meta-reaction
    """
    reactant_species_combination_list = []
    base_species_order = []

    for reactant in reaction.reactants:
        flag_absent_reactant = False
        for _ in range(reactant['stoichiometry']):

            species_for_reactant = []
            base_species_order.append((reactant['object'], reactant['label']))

            for species in meta_species_in_model:
                if reactant['object'] in species.get_references():
                    species_for_reactant.append({'object': species,
                                                 'characteristics': reactant['characteristics']})
                    flag_absent_reactant = True

            if not flag_absent_reactant:
                simlog.error(f'Species {reactant["object"]} or any inheritors were not found in model \n'
                             f'For reaction {reaction} \n'
                             f'Please add the species or remove the reaction')

            reactant_species_combination_list.append(species_for_reactant)

    return base_species_order, reactant_species_combination_list


def construct_rate_function_arguments(rate_function, reaction):
    rate_function_arguments = str(signature(rate_function))

    black_list = ['*', '=']
    if any(i in rate_function_arguments for i in black_list):
        simlog.error(f'Rate arguments must not contain = or *. \n'
                     f'Error in reaction {reaction}. \n'
                     f'Error in rate function {rate_function} in signature {str(signature(rate_function))}')

    rate_function_arguments = str(rate_function_arguments).replace('(', '')
    rate_function_arguments = str(rate_function_arguments).replace(')', '')
    rate_function_arguments = str(rate_function_arguments).replace(' ', '')
    rate_function_arguments = rate_function_arguments.split(',')
    return rate_function_arguments


def create_all_reactions(reactions, meta_species_in_model,
                         ref_characteristics_to_object,
                         type_of_model, dimension):
    """
        This function creates all reactions
        Returns the reactions_for_sbml and parameters_for_sbml dictionary
        Those will be used by another module to create the SBML file

        :param reactions: (meta-reaction objects) reactions objects constructed by the meta_class module
        :param meta_species_in_model: (list) list of meta-species in model
        :param ref_characteristics_to_object: (dict) Characteristics as keys objects as values
        :param type_of_model: (str) stochastic or deterministic
        :param dimension: (int) model dimension 1D, 2D, 3D, .....

        :returns: reactions_for_sbml (dict) = dictionary with all reactions that will be added to the sbml model file,
        parameters_for_sbml (dict) = parameters for the sbml model file
    """
    reactions_for_sbml = {}
    parameters_for_sbml = {}

    check_for_invalid_reactions(reactions, ref_characteristics_to_object)

    for reaction in reactions:

        base_species_order, reactant_species_combination_list = get_involved_species(reaction, meta_species_in_model)

        for combination_of_reactant_species in iterator_for_combinations(reactant_species_combination_list):

            reactant_species_string_combination_list = \
                construct_reactant_structures(combination_of_reactant_species, ref_characteristics_to_object)

            for reactant_string_list in iterator_for_combinations(reactant_species_string_combination_list):

                product_object_list = construct_product_structure(reaction)
                order_structure = construct_order_structure(base_species_order, reactant_string_list)

                if reaction.order is None:
                    product_species_species_string_combination_list = Default(order_structure, product_object_list,
                                                                              meta_species_in_model,
                                                                              ref_characteristics_to_object)
                else:
                    product_species_species_string_combination_list = reaction.order(order_structure,
                                                                                     product_object_list,
                                                                                     meta_species_in_model,
                                                                                     ref_characteristics_to_object)

                for product_string_list in iterator_for_combinations(product_species_species_string_combination_list):

                    reactant_strings = ['_dot_'.join([reactant[0].get_name()] + reactant[1:])
                                        if len(reactant) > 1 else reactant[0].get_name()
                                        for reactant in reactant_string_list]

                    reaction_rate_arguments = None
                    if callable(reaction.rate):
                        reaction_rate_arguments = construct_rate_function_arguments(reaction.rate, reaction)

                    rate_string = fr.extract_reaction_rate(combination_of_reactant_species,
                                                           reactant_strings
                                                           , reaction.rate, type_of_model,
                                                           dimension, reaction_rate_arguments)
                    if rate_string == 0:
                        continue

                    reactions_for_sbml['reaction_' + str(len(reactions_for_sbml))] = \
                        construct_single_reaction_for_sbml(reactant_strings,
                                                           product_string_list,
                                                           rate_string)

    return reactions_for_sbml, parameters_for_sbml


if __name__ == '__main__':
    pass



"""
    This model is responsible for storing example parameters
    It's used by the simulation to check which parameters exist or not.
    If a new parameter is added, they key must be added here as an example.
    Otherwise the simulation object will throw a compilation error, telling the user that the parameter is not
    supported,
"""


def get_example_parameters():
    example_parameters = {
        "__comment_1": "Model parameters - Repetitions only for stochastic",

        "volume": 1,
        "repetitions": 3,
        "level": 0,

        "__comment_2": "basiCO parameters",

        "simulation_method": "stochastic",
        "method": None,
        "start_time": 0,
        "duration": None,
        "r_tol": 1e-8,
        "a_tol": 1e-10,
        "seeds": [1, 2, 3],
        "step_size": 1,

        "__comment_3": "OUTPUT",

        "unit_x": "year",
        "unit_y": "nanomolar",
        "output_concentration": True,
        "output_dir": "",
        "output_event": False,
        "output_file": None,
        "save_data": False,
        "plot_data": True

    }
    return example_parameters



import plotly.express as px
from mobspy import *


"""
    Here we have a NOR_GATE 
    There are two possible repressors for the Promoter A and B
    If any of them bind to the Promoter the protein can no longer be expressed
"""


def NOR_GATE(A_conc, B_conc):
    # Here we define the Protein to be produced, the Promoter that will act as the gate
    # A and B are the inputs any of them can inactivate the Promoter so they inherit from Repressor
    Repressor, Promoter, Protein = BaseSpecies(3)
    Repressor + Promoter.active >> Promoter.inactive [0.5]
    A, B = New(Repressor, 2)
    Promoter >> Promoter + Protein [lambda promoter: 1 if promoter.active else 0]
    Protein >> Zero [2]

    Promoter(100)
    A(A_conc), B(B_conc)
    MySim = Simulation(A | B | Promoter | Protein)
    MySim.duration = 10
    MySim.save_data = False
    MySim.plot_data = False
    MySim.run()
    return MySim.results['data']['Protein']['runs'][0][-1]


heatmap = []
for a in [0, 25, 50, 75, 100]:
    heatmap_line = []
    for b in [0, 25, 50, 75, 100]:
        output = NOR_GATE(a, b)
        heatmap_line.append(output)
    heatmap.append(heatmap_line)

for line in heatmap:
    print(line)

fig = px.imshow(heatmap, x=[0, 25, 50, 75, 100], y=[0, 25, 50, 75, 100], labels=dict(x='b', y='a'))
fig.show()


from mobspy import *
'''
    MobsPy allows Meta-Species to inherit reactions from other Meta-Species
    thus reducing the number of reactions needed to be programmed. 
    In this example we will go through the basics of inheritance.
'''

# We start by defining the BaseSpecies we will use.
Duplicator, Mortal, Eater, Food = BaseSpecies(4)

# Next, we define the reactions for each of the BaseSpecies.
# Zero represents no species.
Duplicator >> 2*Duplicator [1]
Eater + Food >> Eater [1]
Mortal >> Zero [1]

# Multiplication:
#   To define a bacteria species, we will use multiplication.
#   In MobsPy multiplication allows for the inheritance of reactions
#   from the factor-species.
# For example, we would like to define a Bacteria species that is also a Duplicator,
# an Eater, and a Mortal.
# This is done via
Bacteria = Duplicator*Eater*Mortal

# Another way to inherit is using the New call.
# Now New_Bacteria inherits from Bacteria, which means it is also a Duplicator, an Eater, and a Mortal.
New_Bacteria = New(Bacteria)

# The New operator may be used to generate several species at once as
# in this example.
Glucose, Amino = New(Food, 2)

# MobsPy will create all possible combinations for inheritance when defining the model,
# which means that both Bacteria and New_Bacteria will eat Glucose and Amino,
# although this was defined after the Bacteria species.

# Finally, we define the species we want to simulate.
# Since Duplicator, Eater, Mortal, and Food were just defined to construct the other meta-species,
# we don't pass them to Simulation.
MySim = Simulation(Bacteria | New_Bacteria | Glucose | Amino)

# Compile returns a model string that defines the resulting model.
# This allows one to see all the reactions, species, mappings, and parameters.
# The result is available in 2_Reaction_Inheritance_Model.txt
print(MySim.compile())




from mobspy import *
""" 
    In this model we will explain the born species. This is how we refer to species that appear in the products
    but do not appear in the reactants
"""

# Let's look at the following mobspy model:
A = BaseSpecies(1)
B, C = New(A, 2)
B.b1, B.b2, C.c1, C.c2


Zero >> 2 * A[1]
# In the reaction above we add the meta-species A from which B, C inherit from
# However, A does not appear in the reactants
# MobsPy them does not have a paring, so no meta-species to pick from the reactant
# Here only two reactions are defined from this meta-reaction
# Which are the default states of all meta-species

MySim = Simulation(B | C)
print(MySim.compile())

# If one wants the born species meta-reaction to create reactions for all the species in the set,
# he can use the All default order in the following way:

A = BaseSpecies(1)
B, C = New(A, 2)
B.b1, B.b2, C.c1, C.c2
Zero >> 2 * A[1]
MySim = Simulation(B | C)
MySim.default_order = All
print(MySim.compile())


import sys, os
import matplotlib.pyplot as plt
from mobspy import *

# Model definition
"""
    This model is a simple single repressor
"""

Chemical, Promoter, Protein = BaseSpecies(3)

Rev[Chemical + Promoter.inactive >> Promoter.active][2, 1]
Promoter >> Promoter + Protein [lambda promoter: 1 if promoter.inactive else 0]
Protein >> Zero [0.1]
Promoter.inactive(100), Chemical(1000)

# Simulation
MySim = Simulation(Chemical | Promoter | Protein)
MySim.save_data = False
MySim.plot_data = False
MySim.duration = 200
MySim.run()
MySim.plot_deterministic(Protein)


from mobspy import *
"""
    MobsPy can also add characteristics to species 
    and construct reactions based on this change of characteristics by a querying mechanism
    In this example we describe the Query Mechanism in detail 
"""

# First we define 2 BaseSpecies.
Age, Color = BaseSpecies(2)

# Characteristics:
#   We add characteristics to them using the dot operator.
#   Characteristics can be added explicitly or implicitly.
#   For instance the Color  meta-species uses explicit adding.
Color.blue, Color.red, Color.yellow
#   Alternatively, they can be added implicitly within the reaction definition.
#   An example for the latter is the Age  meta-species.
Age.young >> Age.old [1]


# We next multiply Age and Color.
# The new  meta-species inherits the characteristics from Color and Age.
# It inherits them as a two dimensional space where the characteristics sets are orthogonal.
# The Dummy meta-species represents all the possible combinations of characteristics from the meta-species
# it inherits from, separated by a dot.
# As an example, Dummy now represents the following species:
#   Dummy.blue.young, Dummy.blue.old, Dummy.red.young, Dummy.red.old, ....
Dummy = Age*Color

# Queries:
#   We can now query the meta-species that Dummy represents.
#   This allows one to define a reactions for a subset of species in Dummy
#   that match the query.
#   In the meta-reaction bellow, only the species from Dummy with the old characteristic
#   are assigned the reaction
Dummy.old >> Zero [1]

# We can also query over multiple characteristics with the order being irrelevant.
Dummy.young.red >> Dummy.blue [1]

# To generate the compiled string, we execute
MySim = Simulation(Dummy)
print(MySim.compile())
# Results are available in 3_Queries.txt



"""
    This model is responsible for storing MobsPy default parameters
"""
from mobspy.parameter_scripts.parameter_reader import read_json


# This was just created to avoid potential directory compatibilities
def get_default_parameters():
    default_parameters = {
        "__comment_1": "Model parameters - Repetitions only for stochastic",

        "volume": 1,
        "repetitions": 1,
        "level": 3,

        "__comment_2": "basiCO parameters",

        "simulation_method": "deterministic",
        "method": None,
        "start_time": 0,
        "duration": 60,
        "r_tol": 1e-8,
        "a_tol": 1e-10,

        "__comment_3": "OUTPUT",

        "output_dir": "outputs/",
        "output_file": None,
        "output_event": False,
        "unit_x": None,
        "unit_y": None,
        "output_concentration": False,
        "save_data": False,
        "plot_data": True,

        "_continuous_simulation": False,
        "_end_condition": None,
        "_with_event": False
    }
    return default_parameters



"""
    This model is responsible for storing example plotting parameters
    It's used by the simulation to check which parameters exist or not
"""

def get_example_plot_parameters():
    example_parameters = {
        "output_dir": "",
        "logscale": ["X", "Y"],
        "xlim": [0, 1],
        "ylim": [0, 1e3],
        "figsize": [1.5, 1.5],
        "no_title": True,
        "time_series": [0],
        'linestyle':'-',
        'linewidth':0.1,
        'label':'test_label',
        'runs':[0],
        'fill_between':False,
        'frameon':False,
        'title':'Test_Plot',
        'annotations':['Test_note', 0, 0],
        'pad':1,
        'dpi':3,
        "plots":False,
        "species_to_plot": None,

        "figures": [{
            "plots": [{
                "species_to_plot": ["A"]
            }]
        }],

        # These parameters are not supported by plot_raw
        # Their only function is to alter the plot labels so please use that parameter
        # They are only listed here to avoid printing out errors during parameter checks
        "unit_x": '',
        "unit_y": '',
        "output_concentration": False,
        'simulation_method':'stochastic'
    }
    return example_parameters




"""
    This module is responsible for dealing with units in the modules directory level
    MobsPy standard units are Decimeter (Liter) - Second - Count
"""
from pint import UnitRegistry, Quantity
from scipy.constants import N_A
from copy import deepcopy
import os, sys
import mobspy.simulation_logging.log_scripts as simlog
import mobspy.modules.unit_handler as uh


def convert_rate(quantity, reaction_order, dimension):
    """
        This function converts the rate from the users given unit to MobsPy standard units

        :param quantity: (int, float, Quantity) If it is a quantity object convert, otherwise it remains the same
        :param reaction_order: (int) number of reactants in the reaction, to check if the rate is in the correct unit
        :param dimension: (int) model's dimension (1D, 2D, 3D, ... )

        :param quantity: (int, float) converted unit into MobsPy standard units
    """
    volume_power = reaction_order - 1
    converted_quantity = deepcopy(quantity)

    # Dimension arrives as none but there is a rate
    if dimension is None and isinstance(quantity, Quantity) and reaction_order > 1:
        dimension = uh.extract_length_dimension(str(quantity.dimensionality), dimension,
                                                reaction_order)

    if isinstance(quantity, Quantity):
        try:
            if volume_power <= 0:
                if '[substance]' in str(quantity.dimensionality):
                    converted_quantity.ito(f'moles/seconds')
                    return converted_quantity.magnitude, dimension
                else:
                    converted_quantity.ito(f'1/seconds')
                    return converted_quantity.magnitude, dimension
            else:
                if '[substance]' in str(quantity.dimensionality):
                    converted_quantity.ito(f'decimeters ** {dimension*volume_power}/(moles ** {volume_power} * seconds)')
                    return converted_quantity.magnitude / (N_A ** volume_power), dimension
                else:
                    converted_quantity.ito(f'decimeters ** {dimension*volume_power}/seconds')
                    return converted_quantity.magnitude, dimension
        except Exception as e:
            simlog.error(str(e) + '\n' +
                         f'Problem converting rate {quantity} \n'
                         f'Is the rate in the form [volume]**{volume_power}/[time]?')

    else:
        return quantity, dimension


def convert_counts(quantity, volume, dimension):
    """
        This function converts the counts from the users given unit to MobsPy standard units. It also converts
        concentrations into counts

        :param quantity: (int, float, Quantity) If it is a quantity object convert, otherwise it remains the same
        :param volume: (int, float) volume in liters (converted beforehand)
        :param dimension: (int) model's dimension (1D, 2D, 3D, ... )

        :return: converted_quantity (int, float) = converted unit into MobsPy standard units
    """
    converted_quantity = deepcopy(quantity)

    if isinstance(quantity, Quantity):
        if '[length]' not in str(quantity.dimensionality) and '[substance]' not in quantity.dimensionality:
            simlog.error(f'The assigned quantity {quantity} is neither a count or concentration')

        try:
            if '[substance]' in quantity.dimensionality:
                if '[length]' in str(quantity.dimensionality):
                    converted_quantity.ito(f'moles/(decimeter ** {dimension})')
                    converted_quantity = converted_quantity*volume

                converted_quantity = converted_quantity.magnitude * N_A
            else:
                if '[length]' in str(quantity.dimensionality):
                    converted_quantity.ito(f'1/(decimeter ** {dimension})')
                    converted_quantity = converted_quantity*volume
                converted_quantity = converted_quantity.magnitude
        except Exception as e:
            simlog.error(str(e) + '\n' +
                         f'Problem converting rate {quantity} \n'
                         f'Is it really a count or concentration?')
    return converted_quantity


def check_dimension(dimension, value):
    """
        Checks for dimension consistency. It "stores" the first dimension it was given by returning it

        :param dimension: (int) model's dimension (1D, 2D, 3D ...)
        :param value: (int) dimension value being analysed

        :raise simlog.error: If dimensions are not consistent through the given units (units in 1D with 2D mixed)

        :return: dimension (int) = model's dimension (1D, 2D, 3D ...)
    """
    if dimension is None:
        dimension = int(value)
    else:
        if dimension != int(value):
            simlog.error('The dimensions are not consistent. There are at least two units given for different '
                         'dimension models')
    return dimension


def extract_length_dimension(unit_string, dimension, reaction_order=None):
    """
        Extracts the volume dimension from a Quantity object from Pint

        :param unit_string: (str) unit in str format
        :param dimension: (int) model's dimension (1D, 2D, 3D ...)
    """
    temp_list = unit_string.split()
    try:
        position = temp_list.index('[length]')
    except ValueError:
        position = -1

    if position == -1:
        return False
    if temp_list[position + 1] == '**':
        if reaction_order is None:
            dimension = check_dimension(dimension, temp_list[position + 2])
        else:
            temp_int = int(int(temp_list[position + 2])/(reaction_order - 1))
            dimension = check_dimension(dimension, temp_int)
    else:
        dimension = check_dimension(dimension, 1)
    return dimension


def convert_volume(volume, dimension):
    """
        Converts volume to decimetre**dimension

        :param volume: (int, float, Quantity) volume used in simulation

        :return: the converted volume in MobsPy units
    """
    if isinstance(volume, Quantity):
        dimension = extract_length_dimension(str(volume.dimensionality), dimension)
        volume = volume.to(f'decimeter ** {dimension}').magnitude
        return volume
    else:
        return volume


def convert_time(time):
    """
        Converts time into seconds

        :param time: (int, float, Quantity) any time used
    """
    if isinstance(time, Quantity):
        if str(time.dimensionality) == '[time]':
            return time.to('second').magnitude
    else:
        return time


if __name__ == '__main__':
    pass



import itertools


def characteristics_dictionary(characteristics, characteristics_to_object):
    """
        This function constructs a dictionary that leads from the species string characteristics to their respective
        object. This structure allows to easily find where the characteristics in a species string are locates

        :param characteristics: (set) of characteristics
        :param characteristics_to_object: (dict) with characteristics as keys and their respective base meta-species
        object as value
    """
    object_to_characteristic = {}
    for characteristic in characteristics:
        if '$' not in characteristic:
            object_to_characteristic[characteristics_to_object[characteristic]] = characteristic
    return object_to_characteristic


def construct_species_char_list(spe_object, characteristics, characteristics_to_object, symbol=None):
    """
        This function constructs a list in the format ['species_name', 'char1', 'char2', ...]. It generetes this list
        for a given meta-species and the specified characteristics. Values of characteristics not specified in a
        particular position are replaced by their default value. If a symbol is given it generates a string from
        the list using the symbol to join it.

        :param spe_object: meta-species object to be used
        :param characteristics: (set) of characteristics given
        :param characteristics_to_object: (dict) with characteristics as keys and their respective base meta-species
        object as value
        :param symbol: (str) usually . or _dot_, connects the elements from the list using the specified symbol
    """
    if characteristics == 'std$':
        characteristics = set()

    ordered_references_list = spe_object.get_ordered_references()

    objects_to_characteristic = characteristics_dictionary(characteristics, characteristics_to_object)

    species_char_list = [spe_object]
    for obj in ordered_references_list:
        if obj in objects_to_characteristic:
            species_char_list.append(objects_to_characteristic[obj])
        else:
            species_char_list.append(obj.first_characteristic)

    if symbol is not None:
        species_char_list = symbol.join([spe_object.get_name()] + species_char_list [1:]) \
            if len(species_char_list ) > 1 else spe_object.get_name()

    return species_char_list


def construct_all_combinations(spe_object, characteristics, characteristics_to_object, symbol=None):
    """
       This function constructs all possible list in the format ['species_name', 'char1', 'char2', ...] using
       all combinations of characteristics from the vector coordinates not used in the characteristics specified
       in the function argument. If a symbol is given it generates a string from the list using the symbol to join it.

       :param spe_object: meta-species object to be used
       :param characteristics: (set) of characteristics given
       :param characteristics_to_object: (dict) with characteristics as keys and their respective base meta-species
       object as value
       :param symbol: (str) usually . or _dot_, connects the elements from the list using the specified symbol
    """

    if characteristics == 'std$':
        characteristics = set()

    spe_object.order_references()
    ordered_references_list = spe_object.get_ordered_references()

    objects_to_characteristic = characteristics_dictionary(characteristics, characteristics_to_object)

    list_of_all_possibilities = [[spe_object]]
    for obj in ordered_references_list:
        if obj in objects_to_characteristic:
            list_of_all_possibilities.append([objects_to_characteristic[obj]])
        else:
            list_of_all_possibilities.append(list(obj.get_characteristics()))

    to_return = []
    for i in itertools.product(*list_of_all_possibilities):
        if symbol is not None:
            to_return.append(symbol.join([spe_object.get_name()] + list(i)[1:])
                             if len(i) > 1 else spe_object.get_name())
        else:
            to_return.append(list(i))

    return to_return






"""
    Main MobsPy module. It stocks the Simulation class which is responsible for simulating a Model
"""
from copy import deepcopy
from contextlib import contextmanager
from mobspy.modules import meta_class
from mobspy.modules import event_functions
from mobspy.sbml_simulator import run
import mobspy.simulation_logging.log_scripts as simlog
from mobspy.modules.meta_class import *
from mobspy.parameter_scripts import parameter_reader as pr
from mobspy.parameters.default_reader import get_default_parameters
from mobspy.parameters.example_reader import get_example_parameters
from mobspy.plot_params.default_plot_reader import get_default_plot_parameters
import mobspy.sbml_simulator.builder as sbml_builder
import mobspy.sbml_simulator.run as sbml_run
import mobspy.plot_scripts.default_plots as dp
import mobspy.data_handler.process_result_data as dh
from mobspy.modules.set_counts_module import set_counts
from mobspy.data_handler.time_series_object import MobsPyTimeSeries
import json
import os
import inspect
import mobspy.modules.unit_handler as uh
from pint import UnitRegistry

# u is reserved for units
u = UnitRegistry()


class Simulation:

    # Event Implementation
    @classmethod
    def event_compilation_error(cls):
        simlog.error('The event condition did not compile.\n'
                     'Please make sure it follows the following format:\n'
                     'For simple conditions - if C1 \n'
                     'For and based condition - if (C1) & (C2)\n'
                     'For or based conditions - if (C1) & (C2)\n'
                     'Please include the parentheses')

    def event_context_finish(self):
        self._event_time = 0
        for meta_species in self._species_to_set:
            meta_species.reset_simulation_context()
        self._species_to_set = set()
        self._context_not_active = True

    def event_context_add(self, time, trigger):

        event_data = {'event_time': time, 'event_counts': list(self.current_event_count_data),
                      'trigger': trigger}

        self.current_event_count_data = []
        self.pre_number_of_context_comparisons = self.number_of_context_comparisons
        self.number_of_context_comparisons = 0

        if len(event_data['event_counts']) != 0:
            self.total_packed_events.append(event_data)

        self.event_context_finish()

    def event_context_initiator(self):

        # Set context in all meta-species
        if len(self._species_to_set) == 0:
            for i in range(len(inspect.stack())):
                local_names = inspect.stack()[i][0].f_locals
                global_names = inspect.stack()[i][0].f_globals
                for key, item in global_names.items():
                    if isinstance(item, Species):
                        self._species_to_set.add(item)
                for key, item in local_names.items():
                    if isinstance(item, Species):
                        self._species_to_set.add(item)

            for meta_species in self._species_to_set:
                meta_species.set_simulation_context(self)
        else:
            pass

    def _event_handler(self):
        if self._context_not_active:
            self._context_not_active = False
            self.__dict__['parameters']['_with_event'] = True
            self.event_context_initiator()
        else:
            simlog.error('MobsPy does not support multiple context calls')

    @contextmanager
    def event_condition(self, trigger, delay=0):
        try:
            code_line = inspect.stack()[2].code_context[0][:-1]
            if '==' in code_line:
                simlog.error('Equality comparison operator (==) not allowed for MobsPy events \n' +
                             'Please use (A <= n) & (A >= n) if necessary', stack_index=3)
            if type(trigger) == bool or type(trigger) == float or type(trigger) == int:
                simlog.error(f'MobsPy has received an invalid trigger type: {type(trigger)} \n' +
                             f'Please make sure you are not using the operator == for creating event conditions \n'
                             , stack_index=3)
            self._conditional_event = True
            self._event_handler()
            yield 0
        finally:
            delay = uh.convert_time(delay)
            self._conditional_event = False
            self.event_context_add(delay, trigger)

    @contextmanager
    def event_time(self, time):
        try:
            self._event_handler()
            yield 0
        finally:
            time = uh.convert_time(time)
            self.event_context_add(time, 'true')

    def __init__(self, model, names=None, parameters=None, plot_parameters=None):
        """
            Constructor of the simulation object

            Parameters:
            :param model: (List_Species object) Meta-Species object for modeling
            :param names: (dict) names of the meta-species in globals() format. If none it uses the variable names
            :param parameters: (dict) Simulation object parameters. If none takes default parameters
            :param plot_parameters: (dict) Parameters for plotting. If none takes default
        """

        # Event Variable Definitions
        self._species_to_set = set()
        self._event_time = 0
        self.previous_trigger = None
        self.current_event_count_data = []
        self.total_packed_events = []
        self.number_of_context_comparisons = 0
        self.pre_number_of_context_comparisons = 0
        self._list_of_models = []
        self._list_of_parameters = []
        self._context_not_active = True
        self._assigned_species_list = []
        self._conditional_event = False
        self._end_condition = None

        # HERE FABRICIO
        # Get all names
        #if names is None:
        #    local_names = inspect.stack()[1][0].f_locals
        #    global_names = inspect.stack()[1][0].f_globals
        #    names = {}
        #    for key, item in global_names.items():
        #        names[key] = item
        #    for key, item in local_names.items():
        #        names[key] = item

        self.model = model
        self.names = names

        if not isinstance(model, Species) and not isinstance(model, List_Species):
            simlog.error('Model must be formed only by Species objects or List_Species objects \n'
                         f'Model type {type(model)} and it is {model}')

        self.orthogonal_vector_structure = mcu.create_orthogonal_vector_structure(model)

        # Get all meta - reactions
        self._reactions_set = set()
        for spe_object in self.model:
            for reference in spe_object.get_references():
                self._reactions_set = self._reactions_set.union(reference.get_reactions())

        self._species_counts = []
        for spe_object in self.model:
            for count in spe_object.get_quantities():
                self._species_counts.append({'object': spe_object, 'characteristics': count['characteristics'],
                                             'quantity': count['quantity']})
            spe_object.reset_counts()

        if not parameters:
            self.parameters = get_default_parameters()

        if not plot_parameters:
            self.plot_parameters = get_default_plot_parameters()

        # Other needed things for simulating
        self.results = {}
        self.default_order = Default

        self._species_for_sbml = None
        self._reactions_for_sbml = None
        self._parameters_for_sbml = None
        self._mappings_for_sbml = None
        self._events_for_sbml = None
        self.model_string = ''

    def compile(self, verbose=True):
        """
            Compiler method that calls the Compiler class in the modules directory

            :param verbose: (bool) = print or not the results of the compilation
        """
        simlog.global_simlog_level = self.parameters['level']
        simlog.debug('Compiling model')

        pr.parameter_process(self.parameters)
        if self.parameters['method'] is not None:
            self.parameters['simulation_method'] = self.parameters['method']

        if self.parameters['simulation_method'].lower() == 'deterministic':
            self.plot_parameters['simulation_method'] = 'deterministic'
        elif self.parameters['simulation_method'].lower() == 'stochastic':
            self.plot_parameters['simulation_method'] = 'stochastic'

        # Pass end condition to dict parameters - It is stored outside of parameters to the parameters serializable
        # However, it is necessary for the compilation so it is passed as a parameter
        self.parameters['_end_condition'] = self._end_condition

        self._species_for_sbml, self._reactions_for_sbml, \
        self._parameters_for_sbml, self._mappings_for_sbml, \
        self.model_string, self._events_for_sbml, self._assigned_species_list = \
            Compiler.compile(self.model,
                             reactions_set=self._reactions_set,
                             species_counts=self._species_counts,
                             orthogonal_vector_structure=self.orthogonal_vector_structure,
                             volume=self.parameters['volume'],
                             type_of_model=self.parameters[
                                 "simulation_method"],
                             verbose=verbose,
                             event_dictionary=self.total_packed_events,
                             continuous_sim=self.parameters['_continuous_simulation'],
                             ending_condition=self.parameters['_end_condition'])

        # The volume is converted to the proper unit at the compiler level
        self.parameters['volume'] = self._parameters_for_sbml['volume'][0]
        self.mappings = deepcopy(self._mappings_for_sbml)

        # Set common parameters for plot and simulation
        self.plot_parameters['unit_x'] = self.parameters['unit_x']
        self.plot_parameters['unit_y'] = self.parameters['unit_y']
        self.plot_parameters['output_concentration'] = self.parameters['output_concentration']

        self.all_species_not_mapped = {}
        for key in self._species_for_sbml:
            self.all_species_not_mapped[key.replace('_dot_', '.')] = self._species_for_sbml[key]

        self._list_of_models += [{'species_for_sbml': self._species_for_sbml,
                                  'parameters_for_sbml': self._parameters_for_sbml,
                                  'reactions_for_sbml': self._reactions_for_sbml,
                                  'events_for_sbml': self._events_for_sbml,
                                  'species_not_mapped': self.all_species_not_mapped,
                                  'mappings': self.mappings,
                                  'assigned_species': self._assigned_species_list}]

        self._list_of_parameters = [self.parameters]

        if self.model_string != '':
            return self.model_string

    def run(self):
        """
            Runs the simulation by colling the models in the sbml_simulator directory.
            Compiles the model if it was not yet compiled
        """
        # Base case - If there are no events we compile the model here
        if self._species_for_sbml is None:
            self.compile(verbose=False)

        simlog.debug('Starting Simulator')

        unprocessed_data = sbml_run.simulate(self._list_of_parameters, self._list_of_models)

        processed_data = []
        for updl in unprocessed_data:
            processed_data.append(dh.convert_data_to_desired_unit(updl, self.parameters['unit_x'],
                                                                  self.parameters['unit_y'],
                                                                  self.output_concentration,
                                                                  self.parameters['volume']))

        data_dict = {'data': processed_data,
                     'params': self.parameters,
                     'models': self._list_of_models}

        self.results = MobsPyTimeSeries(data_dict)

        if self.parameters['save_data']:
            self.save_data()

        if self.parameters['plot_data']:
            methods_list = [x['simulation_method'] for x in self._list_of_parameters]

            if ('stochastic' in methods_list or 'directedMethod' in methods_list) \
                    and self.parameters['repetitions'] > 1:
                self.plot_stochastic()
            else:
                self.plot_deterministic()

    def save_data(self, file=None):
        """
            Saves the simulation result data to a file in json format

            :param file: (str) name of the file to save the data to. If none a default name is provided
        """
        self._save_data(file=file)

    def _save_data(self, file=None):
        """
            Save results manually into file. Useful for jupyter notebook users

            Parameters
                file (str) = name of the file to create and save JSON data
        """
        if file is None:
            try:
                with open(self.parameters["absolute_output_file"], 'w') as f:
                    json.dump(self.results.to_dict(), f, indent=4)
            except Exception as e:
                simlog.warning("Error saving data. Potential solve: file name parameter")
                simlog.warning(str(e))
        else:
            file += '.json'
            with open(file, 'w') as jf:
                json.dump(self.results.to_dict(), jf, indent=4)

    def _pack_data(self, time_series_data):
        """
            Packs data from multiple simulations or external data into one simulation object

            :param time_series_data: (data in MobsPy format) data to be packed in the simulation object
        """
        self.packed_data.append(time_series_data)

    # Dealing with parameters
    def set_from_json(self, file_name):
        """
            Set simulation parameters from json file

            :param file_name: (str) name of the json file
        """
        with open(file_name) as json_file:
            data = json.load(json_file)
            for key in data:
                self.__setattr__(key, data[key])

    def __setattr__(self, name, value):
        """
            __setattr__ override. For setting simulation parameters using the _dot_ operator

            :param name: (str) name of the parameter to set
            :param value: value of the parameter
        """
        white_list = ['default_order', 'volume', 'model', 'names', 'parameters', 'model_string',
                      'plot_parameters', 'results', '_species_for_sbml',
                      '_reactions_for_sbml', '_parameters_for_sbml', '_mappings_for_sbml', 'mappings',
                      'all_species_not_mapped', 'self._species_for_sbml', 'self._reactions_for_sbml',
                      'self._parameters_for_sbml', 'self._mappings_for_sbml', 'self.model_string',
                      'event_times', 'event_models', 'event_count_dics', '_events_for_sbml',
                      'total_packed_events', 'species_initial_counts', '_species_to_set',
                      '_event_time', 'previous_trigger', 'current_event_count_data',
                      'current_condition', 'current_event_trigger_data',
                      'number_of_context_comparisons', 'pre_number_of_context_comparisons', '_continuous_simulation',
                      'initial_duration', '_reactions_set', '_list_of_models', '_list_of_parameters',
                      '_context_not_active', '_species_counts', '_assigned_species_list', '_conditional_event',
                      '_end_condition', 'orthogonal_vector_structure']

        plotted_flag = False
        if name in white_list:
            self.__dict__[name] = value

        if 'plot_flag' in self.__dict__ and self.__dict__['plot_flag']:
            self.__dict__["plot_parameters"][name] = value
            self.__dict__["plot_flag"] = False
            plotted_flag = True

        if not plotted_flag:
            example_parameters = get_example_parameters()
            if name in example_parameters.keys():
                if name == 'duration':
                    if type(value) == bool:
                        simlog.error(f'MobsPy has received an invalid trigger type: {type(value)} \n' +
                                     f'Please make sure you are not using the operator == for ' +
                                     f'creating event conditions \n'
                                     , stack_index=2)

                if name == 'duration' and isinstance(value, MetaSpeciesLogicResolver):
                    self.__dict__['parameters']['_continuous_simulation'] = True
                    self.__dict__['_end_condition'] = value
                    if 'initial_conditional_duration' not in self.__dict__['parameters']:
                        self.__dict__['parameters']['initial_conditional_duration'] = 1
                else:
                    self.__dict__['parameters'][name] = value
            elif name in white_list:
                pass
            else:
                simlog.error(f'Parameter {name} is not supported', stack_index=2)

    def __getattribute__(self, item):
        if item == 'results' and self.__dict__['results'] == {}:
            simlog.error('The results were accessed before the execution of the simulation', stack_index=2)
        return super().__getattribute__(item)

    def __getattr__(self, item):
        """
            __getattr__ override. For the user to be able to set plot parameters as MySim.plot.parameter
        """
        if item == 'plot':
            self.__dict__['plot_flag'] = True
        else:
            self.__dict__['plot_flag'] = False
        return self

    def configure_parameters(self, config):
        """
            Configure simulation parameters from json file or dictionary

            :param file_name: (str) name of the json file
        """
        self.parameters = self.__config_parameters(config)

    def configure_plot_parameters(self, config):
        """
            Configure plot parameters from json file or dictionary

            :param file_name: (str) name of the json file
        """
        self.plot_parameters = self.__config_parameters(config)

    @staticmethod
    def __config_parameters(config):
        """
            Encapsulation for config_plot and config_parameters
        """
        if type(config) == str:
            if os.path.splitext(config)[1] != '.json':
                simlog.error('Wrong file extension', stack_index=3)
            parameters_to_config = pr.read_json(config)
        elif type(config) == dict:
            parameters_to_config = config
        else:
            simlog.error("Parameters must be python dictionary or json file", stack_index=3)
        return parameters_to_config

    # Plotting encapsulation
    def extract_plot_essentials(self, *species):
        """
            Extract essential information for plotting

            :param species: (meta-species objects) meta-species objects to plot
            :return: species_strings (str) = species strings to be plotted, self.results = data resulting from the
            simulation, self.plot_parameters (dict) = parameters for plotting
        """
        if not species:
            species_strings = set()
            for model in self._list_of_models:
                species_strings = species_strings.union(model['mappings'])
        else:
            species_strings = set()

        for spe in species:
            if isinstance(spe, Species):
                species_strings.add(str(spe))
            elif isinstance(spe, Reacting_Species):
                species_strings.add(str(spe))
            elif type(spe) == str:
                species_strings.add(spe)
            else:
                simlog.error('Only species objects or strings for plotting arguments', stack_index=4)

        return species_strings, self.results, self.plot_parameters

    def plot_stochastic(self, *species):
        """
            Calls stochastic plot. See default_plots module in the plot_scripts directory

            :param species: (str or meta-species objects) list of species to be plotted
        """
        plot_essentials = self.extract_plot_essentials(*species)
        dp.stochastic_plot(plot_essentials[0], plot_essentials[1], plot_essentials[2])

    def plot_deterministic(self, *species):
        """
            Calls deterministic plot. See default_plots module in the plot_scripts directory

            :param species: (str or meta-species objects) list of species to be plotted
        """
        plot_essentials = self.extract_plot_essentials(*species)
        dp.deterministic_plot(plot_essentials[0], plot_essentials[1], plot_essentials[2])

    def plot(self, *species):
        """
            Another way of calling plot_deterministic for simplicity

            :param species: (str or meta-species objects) list of species to be plotted
        """
        self.plot_deterministic(*species)

    def plot_raw(self, parameters_or_file):
        """
            Calls raw plot. See default_plots module in the plot_scripts directory

            :param parameters_or_file: json file name with plot parameter configuration or dictionary with plot
            parameter configuration
        """
        dp.raw_plot(self.results, parameters_or_file)

    def __add__(self, other):
        return SimulationComposition(self, other)

    def generate_sbml(self):
        """
            Generates sbmls strings from the current stored models in the simulation

            "return: to_return (list of str) list of sbml files from all the simulations stored
        """
        to_return = []
        for sbml_data in self._list_of_models:
            to_return.append(sbml_builder.build(sbml_data['species_for_sbml'], sbml_data['parameters_for_sbml'],
                                                sbml_data['reactions_for_sbml'], sbml_data['events_for_sbml']))
        return to_return


class SimulationComposition:

    def _compile_multi_simulation(self):
        for sim1 in self.list_of_simulations:
            for sim2 in self.list_of_simulations:
                if sim1 == sim2:
                    continue

                for spe1 in sim1.model:
                    for spe2 in sim2.model:

                        if spe1.get_name() == spe2.get_name():
                            if spe1.get_all_characteristics() != spe2.get_all_characteristics():
                                simlog.error(f'Species {spe1.get_name()} was modified through simulations. \n' +
                                             f'Although reactions can be removed, the characteristics inherited must'
                                             f'remains the same')

    def __init__(self, S1, S2):
        if isinstance(S1, Simulation) and isinstance(S2, Simulation):
            self.list_of_simulations = [S1] + [S2]
        elif isinstance(S1, SimulationComposition) and isinstance(S2, Simulation):
            self.list_of_simulations = S1.list_of_simulations + [S2]
        elif isinstance(S1, Simulation) and isinstance(S2, SimulationComposition):
            self.list_of_simulations = [S1] + S2.list_of_simulations
        elif isinstance(S1, SimulationComposition) and isinstance(S2, SimulationComposition):
            self.list_of_simulations = S1.list_of_simulations + S2.list_of_simulations
        else:
            simlog.error('Simulation compositions can only be performed with other simulations', stack_index=3)
        self.results = None
        self.base_sim = self.list_of_simulations[0]

    def __add__(self, other):
        return SimulationComposition(self, other)

    # FIX THIS
    def __setattr__(self, name, value):
        white_list = ['list_of_simulations', 'results', 'base_sim']
        broad_cast_parameters = ['level', 'method', 'volume']

        if name == 'duration':
            simlog.error('The durations are to be defined specifically to each simulation and not for the concatenated'
                         ' object. \n'
                         'Please set the durations for each simulation object independently', stack_index=2)

        if name in broad_cast_parameters:
            if name == 'volume':
                for sim in self.list_of_simulations:
                    if sim.__dict__['parameters'][name] != 1:
                        simlog.error('Volumes must be defined only individually for each simulation or once in the '
                                     'concatenated simulation', stack_index=2)
            else:
                for sim in self.list_of_simulations:
                    sim.__dict__['parameters'][name] = value

        if name in white_list:
            self.__dict__[name] = value
        else:
            self.base_sim.__setattr__(name, value)

    def compile(self, verbose=True):
        str = ''
        for sim in self.list_of_simulations:
            str += sim.compile(verbose)
        self._compile_multi_simulation()
        if str != '':
            return str

    def run(self):
        for sim in self.list_of_simulations:
            if sim._species_for_sbml is None:
                sim.compile(verbose=False)
        self._compile_multi_simulation()

        for sim in self.list_of_simulations:
            if sim == self.base_sim:
                continue

            self.base_sim._list_of_models += sim._list_of_models
            self.base_sim._list_of_parameters += sim._list_of_parameters
        self.base_sim.run()
        self.results = self.base_sim.results

    def plot_deterministic(self, *species):
        self.base_sim.plot_deterministic(*species)

    def plot_stochastic(self, *species):
        self.base_sim.plot_stochastic(*species)

    def plot(self, *species):
        self.base_sim.plot(*species)

    def plot_raw(self, parameters_or_file):
        self.base_sim.plot_raw(parameters_or_file)

    def generate_sbml(self):
        return self.base_sim.generate_sbml()


if __name__ == '__main__':
    pass



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt") as fh:
    install_requires = fh.read().strip().split("\n")

with open("mobspy/_version.py") as fh:
    version = fh.readlines()[-1].split()[-1].strip("\"'")

setuptools.setup(
    name="mobspy",
#    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    packages=setuptools.find_packages(include=["mobspy", "mobspy.*"]),
    version=version,
    description="A Query-Based Language for Chemical Reaction Networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ROBACON/mobspy",
    python_requires=">=3.6",
    install_requires=install_requires
)



from mobspy import *
"""
    This model is designed to explain the count assignment process in detail
"""

# Start with the base species
Age, Size = BaseSpecies(2)

# Define some simple reactions
Age.young >> Age.old [1]
Size.small >> Size.medium [1]
Size.medium >> Size.big [1]

# Define a species through multiplication
Thing = Age*Size

# Now the meta-species Thing has 6 species
# If we use the call operator in Things it will assign the count to the default state
# The default state is formed by only the first characteristic that added to each meta-species
# So the count will be assigned to Thing.small.young
Thing(100)

# If one wants to assign a count to another species, just use the dot operator
# Every characteristic not specified will be replaced by the default
Thing.medium(100)
Thing.big.old(100)

MySim = Simulation(Thing)
print(MySim.compile())





import mobspy.simulation_logging.log_scripts as simlog
import inspect
from mobspy.modules.meta_class import List_Species
from pint import Quantity


def set_counts(count_dic):
    """
        Adds counts to meta-species using a given dictionary. Keys from this dictionary can be either meta-species
        objects or strings. Items must be the assign counts to species

        :param count_dic: Dictionary where keys can be either meta-species or strings
        :raise simlog.error: - If a count is assigned to a Reacting_Species with more than one meta-species.
        If there are two species with the same name and a string assignment is performed.
        If the keys are not strings or meta-species.
        If the counts are not Quantities, Floats or Ints.
        If the species was not found in the stack.
        :return: List_Species object. All meta-species that had a count assigned in this dictionary will be returned
        as a List_Species which can be passed as a model to the simulation object
    """
    for key, item in count_dic.items():
        if type(item) == int or type(item) == float or isinstance(item, Quantity):
            continue
        else:
            simlog.error(f'Reactant_species count assignment does not support the type {type(item)}',
                         stack_index=2)

    def find_species():
        found_species = set()

        for i in range(len(inspect.stack())):
            local_names = inspect.stack()[i][0].f_locals
            global_names = inspect.stack()[i][0].f_globals
            for key, item in global_names.items():
                try:
                    if item.is_species() and type(item) != type:
                        found_species.add(item)
                except AttributeError:
                    pass
            for key, item in local_names.items():
                try:
                    if item.is_species() and type(item) != type:
                        found_species.add(item)
                except AttributeError:
                    pass

        return found_species

    for key in count_dic:
        if type(key) == str:
            all_found_species = find_species()

    model = set()
    for key, item in count_dic.items():
        if type(key) == str:
            already_found = False
            str_name = key.split('.')[0]
            str_characteristics = set(key.split('.')[1:])
            for spe in all_found_species:
                if spe.get_name() == str_name and not already_found:
                    already_found = True
                    temp_set = set(str_characteristics)
                    temp_set.discard('all$')
                    if temp_set.issubset(spe.get_all_characteristics()):
                        spe.add_quantities(str_characteristics, item)
                    else:
                        simlog.error('Characteristics not found in species with equal name', stack_index=2)
                    model.add(spe)
                elif spe.get_name() == str_name and already_found:
                    simlog.error(f'There are two different meta-species with the same name. Set_counts cannot resolve',
                                 stack_index=2)
            if not already_found:
                simlog.error(f'Meta-species with the following name {key} not found', stack_index=2)
        else:
            try:
                if key.is_spe_or_reac():
                    if not key.is_species():
                        if len(key.list_of_reactants) != 1:
                            simlog.error('Assignment used incorrectly. Only one species at a time', stack_index=2)
                        model.add(key.list_of_reactants[0]['object'])
                    if key.is_species():
                        model.add(key)
                    key(item)
            except AttributeError:
                simlog.error('Keys must be either meta-species or strings',
                             stack_index=2)

    return List_Species(model)



from mobspy import *
"""
    Restraints
    MobsPy has some restraints for modeling regarding the characteristics combinations
    In this model we will discuss them
"""

try:
    # Here we define two base Species
    A, C = BaseSpecies(2)

    # We them add two characteristics to characteristics to each of them
    A.aaa, A.bbb
    C.aaa, C.ccc

    # And add them to the simulation object
    MySim = Simulation(A | C)
    MySim.level = 0
    MySim.compile()
except SystemExit:
    pass

# This model results in an error:
#   Each species contains a set of the characteristics added to it directly
#   (i.e., characteristics are not obtained by inheritance).
#   If in a simulation two different meta-species contain the same element,
#   the compiler will throw an error.
#   The error message contains the name of the meta-species and the sets of species they cprrespond to.
#
# To avoid this error, use inheritance to share characteristics, or simply rename the characteristics.
#




"""
    mobspy.data_handler.process_result_data.py

    Handles converting the output data from a simulation into desired-units or concentration
"""
from pint import UnitRegistry, Quantity
from scipy.constants import N_A
from copy import deepcopy

u = UnitRegistry()


def convert_data_to_desired_unit(data, unit_x=None, unit_y=None, output_concentration=False, volume=1):
    """Converts the simulation output data from the MobsPy standard units
    to the desired units specified by the user

    :param data: (dict) resulting data from a MobsPy simulation execution
    :param unit_x: (str) unit that the user desires the time in, defaults to dimensionless (None)
    :param unit_y: (str) unit that the user desires the y axis to be in (Concentration or counts)
    :param output_concentration: (bool) decide if output should be a concentration or count
    :param volume: (float) - volume set as parameter to the simulation
    :return: converted_data - input data converted to the desired units
    :rtype: (dict) Dictionary meta-species as key and run as value
    """
    u = UnitRegistry()
    converted_data = deepcopy(data)

    if unit_x is not None:
        new_time = []
        for time in data['Time']:
            quantity = time*u.seconds
            new_time.append(quantity.to(unit_x).magnitude)
        converted_data['Time'] = new_time

    def multiply_data_by_factor(data, factor):
        for key in data:
            if key == 'Time':
                continue

            converted_data[key] = [count*factor for count in data[key]]

    def convert_data(data, from_unit, to_unit):
        for key in data:
            if key == 'Time':
                continue

            converted_data[key] = [(count*from_unit).to(to_unit).magnitude for count in data[key]]

    if output_concentration:
        multiply_data_by_factor(converted_data, volume ** -1)

    if unit_y is not None:
        if 'mol' in unit_y:
            multiply_data_by_factor(converted_data, N_A ** -1)
            if output_concentration:
                convert_data(converted_data, u.molar, unit_y)
            else:
                convert_data(converted_data, u.moles, unit_y)

    return converted_data






import sys, os
from mobspy import *

"""
    This is the CRISP oscillator from the MobsPy publication
    We start by declaring 3 base species and assigning their characteristics
    We need introduce characteristics here for the queries in other dimensions later
"""
Promoter, dCas, CasBinding = BaseSpecies(3)
Promoter.active, Promoter.inactive, CasBinding.no_cas, CasBinding.cas

"""
    The New function creates a New species from Promoter and CasBinding
    The created species inherit the characteristics from Promoter and CasBinding
    Thus allowing us to reference the characteristics from the old species and also
    add a new set of independent characteristics to them
"""
DNAPro = New(Promoter)
gRNA = New(CasBinding)

"""
    The reason why the New constructor was used is to add the list of different gRNAs and Promoters
    If they are added to the DNAPro or gRNA species, those states become independent from 
    active, inactive and no_cas, cas respectively 
"""
gRNAs_list = ['g1', 'g2', 'g3']
Promoters_list = ['P1', 'P2', 'P3']

rev_rt = (1.8e-3/(u.nanomolar * u.second), 2.3e-2 / u.minute)
Rev[gRNA.no_cas + dCas >> gRNA.cas][rev_rt]
gRNA.no_cas >> Zero[0.0069 / u.second]

Promoter.active >> 2 * Promoter.active[2.3e-2 / u.minute]
Promoter >> Zero[2.3e-2 / u.minute]

"""
    Simple loop through the gRNAs and Promoters for assigning the activation reaction
"""
for prom, grna in zip(Promoters_list, gRNAs_list):
    act_rt = lambda dna: 5 / u.minute if dna.active else 0
    DNAPro.c(prom) >> gRNA.no_cas.c(grna) + DNAPro.c(prom)[act_rt]

"""
    Simple loop for the repression reaction
    Here we use the characteristics of gRNA_rep_list as the different type of gRNAs
"""
gRNA_rep_List = ['g3', 'g1', 'g2']
for prom, grna in zip(Promoters_list, gRNA_rep_List):
    dna_rt1 = 1.2e-2 * u.liter / (u.nanomoles * u.second)
    dna_rt2 = 2.3e-2 / u.minute
    DNAPro.active.c(prom) + gRNA.cas.c(grna) >> DNAPro.inactive.c(prom)[dna_rt1]
    DNAPro.inactive.c(prom) >> 2 * DNAPro.active.c(prom) + gRNA.cas.c(grna)[dna_rt2]

# Rev defines a reversible reaction in both senses
Rev[Zero >> dCas][1 / u.minute, 2.3e-2 / u.minute]

DNAPro.active.P1(1), DNAPro.active.P2(1), DNAPro.active.P3(1)
gRNA.no_cas.g1(0), gRNA.no_cas.g2(40*u.nanomolar), gRNA.no_cas.g3(1*u.nanomolar)
dCas(43*u.nanomolar)

MySim = Simulation(gRNA | DNAPro | dCas)
MySim.volume = 1*u.femtoliter

# 40 nanomolar, 1
# dCas 43nm
MySim.save_data = False
MySim.plot_data = False
MySim.duration = 650*u.hours
MySim.step_size = 100
MySim.unit_x = 'hours'
MySim.run()
MySim.plot_deterministic(gRNA.g1,
                         gRNA.g2,
                         gRNA.g3)




"""
    mobspy.modules.process_result_data.py

    Handles rate_assignment to the reactions:
        The passing of arguments to be executed by the user supplied rate function
        The construction of mass action kinetics
        Handles the different types of rates supplied by the user
"""

import mobspy.modules.meta_class as mc
import mobspy.modules.reaction_construction_nb as rc
import mobspy.simulation_logging.log_scripts as simlog
import mobspy.modules.meta_class as mc
import mobspy.modules.unit_handler as uh
from pint import Quantity


class Bool_Override:
    """
            Just a base class for implementing the . operation in the rate function arguments
            through boolean overriding. It is responsible for returning true when the reactant
            has the specified characteristic when using the dot notation

            :param _stocked_characteristics: (str) stocks the characteristics of the queries performed by the user
            :param species_string: (str) string value from an individual species in MobsPY format
    """

    def __bool__(self):
        """
            The implementation of the .dot operation for rate function arguments
            Returns true when the argument possesses the characteristics
            Bool is called after the .dot operations

            Parameters:
                self

            Returns:
                True if the object contains all the characteristics queried
                False otherwise
        """
        if self.species_string == '$Null':
            return False

        species_string_split = self.species_string.split('_dot_')[1:]
        if all([char in species_string_split for char in self._stocked_characteristics]):
            to_return_boolean = True
        else:
            to_return_boolean = False

        self._stocked_characteristics = set()
        return to_return_boolean


class Specific_Species_Operator(Bool_Override):
    """
        This class creates objects from the species strings from the meta-species to pass them to the rate functions
        as arguments. It uses the Bool_Override class to return true or false to the .dot operation inside the rate
        functions


        :param _stocked_characteristics: (str) stocks the characteristics of the queries performed by the user
        :param species_string: (str) string value from an individual species in MobsPY format
        :param species_object: (Species) Meta-species object which originated the meta-species str
    """

    def __init__(self, species_string, species_object):
        """
            Constructs the object from the species strings from the meta-species to pass them to the rate functions
            as arguments.

            :param species_string: (str) A string from MobsPy meta-species format
            :param species_object: (Species) = Meta-species set for which the species_string is contained in
        """
        self.species_string = species_string
        self._stocked_characteristics = set()
        self._species_object = species_object

    def __getattr__(self, characteristic):
        """
            Stores the characteristics for the boolean query inside the rate function by adding them to the set

            :param characteristic: (str) characteristic being queried
        """
        self._stocked_characteristics.add(characteristic)
        return self

    def __str__(self):
        """
            Returns the species_string from the MobsPy meta-species used in the object construction
        """
        return self.species_string

    def is_a(self, reference):
        """
            This function checks to see if the meta-species the species_string belong to has inherited from the
            parameter reference (reminder: every meta-species inherits from itself)

            :param reference: (Species) Meta-species object
            :return: (bool) True if the meta-species in Specific_Species_Operator has inherited from the reference
            False otherwise
        """
        if not self._stocked_characteristics:
            if reference in self._species_object.get_references():
                return True
            else:
                return False
        else:
            simlog.error('Concatenation of is_a and dot operator not supported. Please use them separately',
                         stack_index=2)

    def add(self, characteristic):
        """
            Adds characteristic to the set of characteristics when checking for all
            referred characteristics by the user

            :param characteristic: (str) characteristic to add to the set
        """
        self._stocked_characteristics.add(characteristic)


def extract_reaction_rate(combination_of_reactant_species, reactant_string_list
                          , reaction_rate_function, type_of_model, dimension, function_rate_arguments=None):
    """
        This function is responsible for returning the reaction rate string for the model construction. To do this it
        does a different action depending on the type of the reaction_rate_function (we consider constants as functions)
        It passes the rate as a string expression in MobsPy standard units

        :param  function_rate_arguments: list of strings of the function rate argument ex:['r1', 'r2', ....]
        :param  combination_of_reactant_species: (list of Species) Meta-species currently being used in this reaction
        :param  reactant_string_list: (list of strings) list of species strings in order they apear in the reaction
        :param  reaction_rate_function: (float, callable, Quantity) rate stored in the reaction object
        :param  dimension: (int) system dimension (for the rate conversion)

        :return: reaction_rate_string (str) the reaction kinetics as a string for SBML
    """

    if type(reaction_rate_function) == int or type(reaction_rate_function) == float or \
            isinstance(reaction_rate_function, Quantity):
        # Function is a constant function number int here
        reaction_rate_function, dimension = uh.convert_rate(reaction_rate_function, len(reactant_string_list),
                                                            dimension)
        if reaction_rate_function == 0:
            return 0
        reaction_rate_string = basic_kinetics_string(reactant_string_list,
                                                     reaction_rate_function, type_of_model)

    elif function_rate_arguments is not None:

        # [''] means that it is a function that takes no arguments (empty signature)
        if function_rate_arguments != ['']:
            arguments = prepare_arguments_for_callable(combination_of_reactant_species,
                                                       reactant_string_list, function_rate_arguments)

            rate = reaction_rate_function(**arguments)
        else:
            rate = reaction_rate_function()

        rate, dimension = uh.convert_rate(rate, len(reactant_string_list), dimension)
        if rate == 0:
            return 0

        if type(rate) == int or type(rate) == float:
            reaction_rate_string = basic_kinetics_string(reactant_string_list,
                                                         rate, type_of_model)
        elif type(rate) == str:
            reaction_rate_string = rate
        elif rate is None:
            simlog.error('There is a reaction rate missing for the following reactants: \n'
                         + str(reactant_string_list))
        else:
            simlog.error(f'The rate function {reaction_rate_function}, returned a non-valid value. \n' +
                         f'Only int, floats and str are accepted')
    elif reaction_rate_function is None:
        simlog.error('There is a reaction rate missing for the following reactants: \n'
                     + str(reactant_string_list))
    elif type(reaction_rate_function) == str:
        return reaction_rate_function
    else:
        simlog.debug(type(reaction_rate_function))
        simlog.error(f'The {type(reaction_rate_function)}, from {reaction_rate_function} is not supported')

    return reaction_rate_string


def basic_kinetics_string(reactants, reaction_rate, type_of_model):
    """
        This constructs the bases for mass-action kinetics. Both for stochastic and deterministic depending on the type
        of model

        :params reactants: (list of str) list of reactants in MobsPy str format
        :params reaction_rate: (float) reaction constant
        :params type_of_model: (str) stochastic or deterministic - rate expressions are different depending on each case

        :return:  kinetics_string (str) mass action kinetics expression for the reaction
    """
    counts = rc.count_string_dictionary(reactants)

    if type_of_model.lower() not in ['stochastic', 'deterministic']:
        simlog.error(f'The following method for simulation {type_of_model} is not supported.\n' +
                     f'Only stochastic or deterministic')

    kinetics_string = ""
    for name, number in counts.items():
        if type_of_model.lower() == 'stochastic':
            kinetics_string += stochastic_string(name, number)
        elif type_of_model.lower() == 'deterministic':
            kinetics_string += deterministic_string(name, number)
        kinetics_string += ' * '

    kinetics_string += str(reaction_rate)

    n = kinetics_string.count('*') - 1
    if n > 0:
        kinetics_string += f' * volume^{-n}'

    return kinetics_string


def stochastic_string(reactant_name, number):
    """
        This function returns the stochastic string expression for mass action kinetics
        For instance the reaction 2A -> 3A would imply A*(A-1)/2
        It only does so for one reactant, so it must be called for all reactants in the reaction

        :params reactant_name: (str) species string involved in the reaction
        :params number: (int) stoichiometry (number of times it appears)

        :return: to_return_string (str) the mass action kinetics string expression for only that species
    """
    to_return_string = ''
    for i in range(number):
        if i == 0:
            to_return_string = reactant_name
        else:
            to_return_string += f' * ({reactant_name} - {i})/{i + 1}'

    return to_return_string


def deterministic_string(reactant_name, number):
    """
        This function returns the deterministic string expression for mass action kinetics
        For instance the reaction 2A -> 3A would imply A*A
        It only does so for one reactant, so it must be called for all reactants in the reaction

        :params reactant_name: (str) species string involved in the reaction
        :params number: (int) stoichiometry (number of times it appears)

        :return: to_return_string (str) the mass action kinetics string expression for only that species
    """
    to_return_string = ''
    for i in range(number):
        if i == 0:
            to_return_string = reactant_name
        else:
            to_return_string += f' * {reactant_name}'
    return to_return_string


def prepare_arguments_for_callable(combination_of_reactant_species, reactant_string_list, rate_function_arguments):
    """
        This function prepares the requested arguments to the rate function for a given reaction by creating objects
        of the Specific_Species_Operator class

        :params combination_of_reactant_species: meta-species involved in the reaction
        :params reactant_string_list: species strings involved in the reaction
        :params rate_function_arguments: arguments received by the rate function
    """
    argument_dict = {}
    if rate_function_arguments is not None:
        i = 0
        for i, (species, reactant_string) in enumerate(zip(combination_of_reactant_species, reactant_string_list)):
            try:
                species = species['object']
                argument_dict[rate_function_arguments[i]] = Specific_Species_Operator(species_string=reactant_string,
                                                                                      species_object=species)
            except IndexError:
                continue

        while len(argument_dict) < len(rate_function_arguments):
            i += 1
            argument_dict[rate_function_arguments[i]] = Specific_Species_Operator('$Null', mc.Zero)

    return argument_dict


if __name__ == '__main__':
    pass



from mobspy import *
"""
    Here we discuss another complex MobsPy feature the .c query.
    This method queries over the value of a stored variable instead of it's name.
"""

# We start by defining the base species that we will use.
Color = BaseSpecies(1)

# Now we define a two lists of colors.
light = ['blue', 'red', 'yellow', 'pink']
dark = ['dark_blue', 'dark_red', 'dark_yellow', 'dark_pink']

# In the loop bellow we can see how the .c works:
#   We want to design transition reactions from the light colors to their dark versions.
#   Here, the variables lc and dc loop through the light colors and dark colors list.
#   At each loop iteration, lc is assigned the value of a color and
#   dc is set to the dark version of that color.
#   The .c method is then used to query for the string stored inside both variables.
#   The .c query can be used together with a standard query. The order does not mater.

for lc, dc in zip(light, dark):
    # With the .c method we now query over the characteristic value inside lc and dc.
    Color.c(lc) >> Color.c(dc) [1]

MySim = Simulation(Color)
print(MySim.compile())




from copy import deepcopy
import inspect
import sys

global_simlog_level = 3


def error(message, stack_index=-1):
    if stack_index > -1:
        code_line = inspect.stack()[stack_index].code_context[0][:-1]
        line_number = inspect.stack()[stack_index].lineno
        message = f'At: {code_line} \n' + f'Line number: {line_number} \n' + message

    if global_simlog_level >= 0:
        print('\033[91m' + 'ERROR: ' + message + '\033[0m', file=sys.stderr)
    sys.exit(1)


def debug(message=''):
    global global_simlog_level
    if global_simlog_level >= 2:
        message_copy = deepcopy(message)
        print(message_copy.replace('_dot_', '.'), file=sys.stderr)
    pass


def warning(message):
    global global_simlog_level
    if global_simlog_level >= 1:
        message_copy = deepcopy(message)
        message.replace('_dot_', '.')
        print('\033[91m' + 'WARNING: ' + message_copy + '\033[0m', file=sys.stderr)


if __name__ == '__main__':
    a = 5
    error(f'Did not found your {a}')



import basico
from joblib import Parallel, delayed
import mobspy.simulation_logging.log_scripts as simlog
import mobspy.sbml_simulator.builder as sbml_builder
from copy import deepcopy


def simulate(list_of_params, models):
    """
        This function coordinates the simulation by calling the necessary jobs
        In the future we hope to implement parallel cluster computing compatibility

        :param list_of_params: (dict) simulation parameters from the text file
        :param models: (list) [{'species_for_sbml':, 'parameters_for_sbml':, 'reactions_for_sbml':,
                            'events_for_sbml':, 'species_not_mapped':, 'mappings':}]
        :return: data (dict) = dictionary containing the resulting data from simulation
    """
    params = list_of_params[0]

    # Run in parallel or sequentially
    # If nothing is specified just run it in parallel
    try:
        if params["jobs"] == 1:
            simlog.debug("Running simulation sequentially")
            jobs = params["jobs"]
        else:
            simlog.debug("Running simulation in parallel")
            jobs = params["jobs"]
    except KeyError:
        simlog.debug("Running simulation in parallel")
        jobs = -1

    # This is necessary to avoid sending meta-species objects to joblib

    dummy_copy = []
    for p in list_of_params:
        dummy_copy.append(p['_end_condition'])
        p['_end_condition'] = None

    data = job_execution(list_of_params, models, jobs)

    simlog.debug("Simulation is Over")
    return data


def job_execution(params, models, jobs):
    """
        This is defined for parallelism purposes
        Uses multiple cores from the processor to execute stochastic simulations

        :param params: (dict) = simulation parameters
        :param models: (list) [{'species_for_sbml':, 'parameters_for_sbml':, 'reactions_for_sbml':,
                            'events_for_sbml':, 'species_not_mapped':, 'mappings':}]
        :param jobs: (int) = number of cores to use, -1 for all available

        :return: parallel_data data of all the individual simulations executed in parallel
    """

    def __single_run(packed):
        i = packed

        added_data = {}
        # This is only need to avoid git errors
        reformatted_data = {}
        for j, (sim_par, model) in enumerate(zip(params, models)):

            # Generate SBML here
            if j > 0:
                sbml_str = __sbml_new_initial_values(reformatted_data, model, sim_par, new_model=True)
            else:
                sbml_str = __sbml_new_initial_values({}, model, sim_par)

            end_condition_not_satisfied = True
            if sim_par['_continuous_simulation']:
                sim_par['duration'] = sim_par['initial_conditional_duration']

            while end_condition_not_satisfied:
                basico_model = basico.model_io.load_model_from_string(sbml_str)
                data = __run_time_course(basico_model, sim_par['duration'], sim_par, i)

                reformatted_data = reformat_time_series(data)

                if sim_par['_continuous_simulation']:
                    if reformatted_data['End_Flag_MetaSpecies'][-1] > 0:
                        reformatted_data = __filter_condition_event_time_data(reformatted_data)
                        end_condition_not_satisfied = False
                    else:
                        sbml_str = __sbml_new_initial_values(reformatted_data, model, sim_par)
                        sim_par['duration'] = 2 * sim_par['duration']
                else:
                    end_condition_not_satisfied = False

                reformatted_data = __remap_species(reformatted_data, model['mappings'], model['species_for_sbml'])
                added_data = __add_simulations_data(added_data, reformatted_data)

        return added_data

    parallel_data = Parallel(n_jobs=jobs)(delayed(__single_run)(i) for i in range(params[0]['repetitions']))

    if not parallel_data:
        simlog.error("Error: The parallel model has not produced an output." +
                     "Try addding ('sequential': True) to parameters")

    return parallel_data


def __run_time_course(basico_model, duration, params, index):
    """
        This function prepares the parameters for the basiCO.run_time_course

        :param basico_model: basico model to be run in the simulation
        :param params: (dict) simulation parameters
        :param index: (int) current index of the run
        :param duration: (int, float) simulation duration in seconds
    """
    params["simulation_method"] = params["simulation_method"].lower()
    if (params['_with_event'] or params['_continuous_simulation']) and params["simulation_method"] == 'stochastic':
        params["simulation_method"] = 'directmethod'

    kargs = {'model': basico_model,
             'method': params["simulation_method"],
             'start_time': params["start_time"],
             'r_tol': params["r_tol"],
             'a_tol': params["a_tol"],
             'output_event': params['output_event']}

    if 'seeds' in params:
        kargs['use_seed'] = True
        kargs['seed'] = params['seeds'][index]

    if 'step_size' in params:
        kargs['automatic'] = False
        kargs['step_number'] = int(params['duration'] / params['step_size'])

    return basico.run_time_course(duration, **kargs)


def reformat_time_series(data):
    """
        Transforms the _dot_ in from the results into .

        :param data: (pd.dataframe) simulation data results from BasiCO
    """
    data_dict = {'Time': data.index.tolist()}

    for key in data:
        data_dict[key.replace('_dot_', '.')] = list(data[key])

    return data_dict


def __filter_condition_event_time_data(data):
    """
        This function filters all the data until the event was triggered. It is used for the conditional duration
        simulations

        :param data: (dict) single run data with species as keys and values
    """
    new_data = {}

    for i, e in enumerate(data['End_Flag_MetaSpecies']):
        if e == 1:
            stop_index = i
            break

    for key in data:
        new_data[key] = data[key][:stop_index + 1]

    return new_data


def __sbml_new_initial_values(data, model, sim_para, new_model=False):
    """
        This function adjusts the new sbml counts to the end of the previous concatenation simulation or conditional
        simulation

        :param data: (dict) single run data with species as keys and values
        :param model: (dict) model to be used in the next simulation
        :param model: (sim_par) parameters to be used in the next simulation
        :param new_model: (bool) is a new model being used or the model is the same as the old one
    """
    species_for_sbml = model['species_for_sbml']

    check_list = ["stochastic", "directmethod"]
    for key in data:
        sbml_key = key.replace('.', '_dot_')
        if sbml_key not in species_for_sbml.keys():
            continue

        if key == 'Time':
            continue
        try:
            # Case of species set
            if sbml_key in model['assigned_species'] and new_model:
                continue
            if sim_para["simulation_method"].lower() in check_list:
                species_for_sbml[sbml_key] = int(list(data[key])[-1])
            else:
                species_for_sbml[sbml_key] = list(data[key])[-1]
        except KeyError:
            pass

    if new_model:
        try:
            species_for_sbml['End_Flag_MetaSpecies'] = 0
        except KeyError:
            pass

    return sbml_builder.build(species_for_sbml, model['parameters_for_sbml'],
                              model['reactions_for_sbml'], model['events_for_sbml'])


def __add_simulations_data(added_data, reformatted_data):
    """
        Adds the data from the new executed simulation to the data stored so far

        :param added_data: (dict) added data so far in the simulation (key: species string - value: run)
        :param reformatted_data: (dict) data return from the simulation after being formatted using the reformat
        data function
    """
    if added_data != {}:
        time_to_add = added_data['Time'][-1]
    else:
        time_to_add = 0
    new_data = {}
    already_added_keys = set()

    for key in added_data:
        new_data[key] = added_data[key]

    for i, time in enumerate(reformatted_data['Time']):
        # Remove the repeated initial value from following simulations.
        # The final value of summed simulations is repeated
        if time == 0 and added_data != {}:
            for key in reformatted_data:
                reformatted_data[key].pop(0)

        # This is for simulations with only one value. The system did not change at all
        try:
            reformatted_data['Time'][i] = reformatted_data['Time'][i] + time_to_add
        except IndexError:
            pass

    for key in added_data:
        if key == 'Time':
            continue
        try:
            new_data[key] = added_data[key] + reformatted_data[key]
            already_added_keys.add(key)
        except KeyError:
            dummy = added_data[key][-1]
            new_data[key] += [dummy for _ in reformatted_data['Time']]

    for key in reformatted_data:
        if key == 'Time':
            continue

        if key in already_added_keys:
            continue
        else:
            if time_to_add != 0:
                new_data[key] = [0 for _ in added_data['Time']]
                new_data[key] = new_data[key] + reformatted_data[key]
            else:
                new_data[key] = reformatted_data[key]

    if time_to_add != 0:
        new_data['Time'] = added_data['Time'] + reformatted_data['Time']
    else:
        new_data['Time'] = reformatted_data['Time']

    return new_data


def __remap_species(data, mapping, species_not_mapped):
    """
        Takes the simulated species (data) and add ones defined by
        mapping (mapping).
        By default, a mapping is a list of species in which case the
        sum is taken: ['a', 'b', ...]
        It also adds species that have been removed from the basico simulation

        Parameters:
        :param data: (dict) Data in MobsPy format
        :param mapping: (dict) = Mappings between species and meta-species
        :param species_not_mapped: (list of str) Species not mapped by BasiCO
    """

    mapped_data = {'Time': data['Time']}
    T = range(len(data['Time']))

    # copy over all unmapped ones
    for k in data.keys():
        mapped_data[k] = data[k]

    dot_species_not_mapped = {}
    for key in species_not_mapped:
        dot_species_not_mapped[key.replace('_dot_', '.')] = species_not_mapped[key]

    # 1st pass with sum mappings
    for group in mapping.keys():

        the_mapping = mapping[group]
        mapped_data[group] = {'runs': []}

        try:
            # check if is a list -> sum
            if type(the_mapping) is list:

                this_run = []
                runs_not_returned_by_basico = {}
                for t in T:
                    mapping_sum = 0
                    for spe in the_mapping:
                        try:
                            mapping_sum = mapping_sum + data[spe][t]
                        except KeyError:
                            mapping_sum = mapping_sum + dot_species_not_mapped[spe]
                            try:
                                runs_not_returned_by_basico[spe] += [dot_species_not_mapped[spe]]
                            except KeyError:
                                runs_not_returned_by_basico[spe] = [dot_species_not_mapped[spe]]
                    this_run.append(mapping_sum)
                for spe in runs_not_returned_by_basico:
                    try:
                        mapped_data[spe] = runs_not_returned_by_basico[spe]
                    except KeyError:
                        mapped_data[spe] = [runs_not_returned_by_basico[spe]]
                mapped_data[group] = this_run

        except IndexError:
            simlog.error(f'run: remap_species: error when remapping "{the_mapping}".' +
                         'Possible fix: All runs must have the same time')

        except TypeError:
            simlog.warning(f'Copasi removes A >> A species from reaction calculations and does not provide an output')
            simlog.warning(f'Please check the output data to see if this is the problem')
            for key in data:
                print(key, data[key])
            exit(1)

    return mapped_data



from mobspy import *

"""
    In this model we discuss rate assignment possibilities in detail.
"""

# We start with the base species
# and add some characteristics:
Position, HandShaker, Lifter, Box = BaseSpecies(4)
Lifter.weak, Lifter.strong
Position.p1, Position.p2, Position.p3

# The rate function arguments work in the following manner:
#   The rate function returns a value based on the characteristics of the reactants.
#   The value may thus differ for the species in the meta-species.
#   A meta-reaction creates a reaction for all combinations of species.
#   The species that are currently being used to construct a reaction
#   are passed as arguments to the rate function.
#
# Now we want to define a box breaking reaction:
#   Here, we wish to say that strong lifters break the box faster.
#   In this case r1 will receive a species from the Lifter meta-species as it is the first
#   to appear in the reaction.
#   If another argument (say, r2) were passed to the function, MobsPy would pass species from the Box meta-species.
#   This is optional, however.
#   Arguments can have any name.
#   The rate function may filter on characteristics: The dot notation returns true iff the species has that
#   characteristic, and false otherwise.
#   We use this to make sure that only the species from the meta-species lifter that have the strong characteristics
#   will receive the higher rate.
Lifter + Box >> Lifter [lambda r1: 10 if r1.strong else 1]
# The value returned by the rate function is interpreted as the reaction rate constant of a
# rate that follows mass-action kinetics.

# The rate function can return strings as well.
# This can be used if kinetics different from mass-action kinetics are to be specified.
Lifter >> Lifter [lambda r1: f'1/(1+{r1})']

# Finally, one may wish to check the value of a certain characteristic to define a rate.
# For example, if two Position species are at the same place, we want them to become a Handshake.
# This can be done by using the call argument with the meta-species the characteristic has been added to.
Position + Position >> HandShaker [lambda r1, r2: 100 if Position(r1) == Position(r2) else 0]

MySim = Simulation(Position | Lifter | Box | HandShaker)
print(MySim.compile())








#!/usr/bin/env python3

"""
    Picks and translates a python model into an
    SBML readable by Copasi. New models must
    be references here as well.
"""


from mobspy.sbml_simulator.SBMLWriter import create_model
import libsbml as sbml


def build(species, parameters, reactions, events):
    """
    Constructs the sbml file for a model from the dictionary syntax for python sbml lib

    :param species: (dict) species as keys and counts as values
    :param parameters: (dict) parameter name and value
    :param reactions: (dict) reaction name and reaction in python sbml writer format
    :param events: (dict) event name and event in python sbml writer format

    :return: sbml_str (str) = string describing the model in sbml format
    """
    doc = create_model(species, parameters, reactions, events)

    # Convert sbml document into a string for basico
    sbml_str = sbml.writeSBMLToString(doc)
    return sbml_str




